from __future__ import annotations

import re
from typing import Any, Callable, Optional

from codegen.models import AST, ImportHelper, PredefinedFn, Program, expr, stmt
from codegen.models.var import DeferredVar
from loguru import logger

from sera.misc import (
    assert_isinstance,
    assert_not_null,
    identity,
    to_camel_case,
    to_pascal_case,
    to_snake_case,
)
from sera.models import (
    Class,
    DataProperty,
    Enum,
    ObjectProperty,
    Package,
    Schema,
    TsTypeWithDep,
)
from sera.typing import is_set

TS_GLOBAL_IDENTS = {
    "normalizers.normalizeNumber": "sera-db.normalizers",
    "normalizers.normalizeOptionalNumber": "sera-db.normalizers",
}


def make_typescript_data_model(schema: Schema, target_pkg: Package):
    """Generate TypeScript data model from the schema. The data model aligns with the public data model in Python, not the database model."""
    app = target_pkg.app

    # mapping from type alias of idprop to its real type
    idprop_aliases = {}
    for cls in schema.classes.values():
        idprop = cls.get_id_property()
        if idprop is not None:
            idprop_aliases[f"{cls.name}Id"] = (
                idprop.get_data_model_datatype().get_typescript_type()
            )

    def clone_prop(prop: DataProperty | ObjectProperty, value: expr.Expr):
        # detect all complex types is hard, we can assume that any update to this does not mutate
        # the original object, then it's okay.
        return value

    def make_normal(cls: Class, pkg: Package):
        """Make a data model for the normal Python data model"""
        if not cls.is_public:
            # skip classes that are not public
            return

        idprop = cls.get_id_property()
        program = Program()
        program.import_(
            f"@.models.{pkg.dir.name}.Draft{cls.name}.Draft{cls.name}", True
        )

        prop_defs = []
        prop_constructor_assigns = []
        deser_args = []

        for prop in cls.properties.values():
            if prop.data.is_private:
                # skip private fields as this is for APIs exchange
                continue

            propname = to_camel_case(prop.name)

            if isinstance(prop, DataProperty):
                tstype = prop.get_data_model_datatype().get_typescript_type()
                for dep in tstype.deps:
                    program.import_(dep, True)

                if idprop is not None and prop.name == idprop.name:
                    # use id type alias
                    tstype = TsTypeWithDep(f"{cls.name}Id")

                if prop.is_optional:
                    # convert type to optional
                    tstype = tstype.as_optional_type()

                deser_args.append(
                    (
                        expr.ExprIdent(propname),
                        PredefinedFn.attr_getter(
                            expr.ExprIdent("data"), expr.ExprIdent(prop.name)
                        ),
                    )
                )
            else:
                assert isinstance(prop, ObjectProperty)
                if prop.target.db is not None:
                    # this class is stored in the database, we store the id instead
                    tstype = TsTypeWithDep(
                        f"{prop.target.name}Id",
                        [
                            f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.name}.{prop.target.name}Id"
                        ],
                    )
                    if prop.cardinality.is_star_to_many():
                        tstype = tstype.as_list_type()
                    elif prop.is_optional:
                        # convert type to optional only if it isn't a list
                        tstype = tstype.as_optional_type()
                    deser_args.append(
                        (
                            expr.ExprIdent(propname),
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("data"), expr.ExprIdent(prop.name)
                            ),
                        )
                    )
                else:
                    # we are going to store the whole object
                    tstype = TsTypeWithDep(
                        prop.target.name,
                        [
                            f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.name}.{prop.target.name}"
                        ],
                    )
                    if prop.cardinality.is_star_to_many():
                        tstype = tstype.as_list_type()
                        deser_args.append(
                            (
                                expr.ExprIdent(propname),
                                PredefinedFn.map_list(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent("data"),
                                        expr.ExprIdent(prop.name),
                                    ),
                                    lambda item: expr.ExprMethodCall(
                                        expr.ExprIdent(
                                            assert_isinstance(
                                                prop, ObjectProperty
                                            ).target.name
                                        ),
                                        "deser",
                                        [item],
                                    ),
                                ),
                            )
                        )
                    else:
                        if prop.is_optional:
                            # convert type to optional only if it isn't a list
                            tstype = tstype.as_optional_type()
                        deser_args.append(
                            (
                                expr.ExprIdent(propname),
                                expr.ExprFuncCall(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent(prop.target.name),
                                        expr.ExprIdent("deser"),
                                    ),
                                    [
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent("data"),
                                            expr.ExprIdent(prop.name),
                                        )
                                    ],
                                ),
                            )
                        )

                for dep in tstype.deps:
                    program.import_(
                        dep,
                        True,
                    )

            prop_defs.append(stmt.DefClassVarStatement(propname, tstype.type))
            prop_constructor_assigns.append(
                stmt.AssignStatement(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent("this"),
                        expr.ExprIdent(propname),
                    ),
                    expr.ExprIdent("args." + propname),
                )
            )

        program.root(
            stmt.LineBreak(),
            (
                stmt.TypescriptStatement(
                    f"export type {cls.name}Id = {idprop.get_data_model_datatype().get_typescript_type().type};"
                )
                if idprop is not None
                else None
            ),
            stmt.LineBreak(),
            lambda ast00: ast00.class_like(
                "interface",
                cls.name + "ConstructorArgs",
            )(*prop_defs),
            stmt.LineBreak(),
            lambda ast10: ast10.class_(cls.name)(
                *prop_defs,
                stmt.LineBreak(),
                lambda ast11: ast11.func(
                    "constructor",
                    [
                        DeferredVar.simple(
                            "args", expr.ExprIdent(cls.name + "ConstructorArgs")
                        ),
                    ],
                )(*prop_constructor_assigns),
                stmt.LineBreak(),
                lambda ast12: ast12.func(
                    "deser",
                    [
                        DeferredVar.simple("data", expr.ExprIdent("any")),
                    ],
                    expr.ExprIdent(cls.name),
                    is_static=True,
                    comment="Deserialize the data from the server to create a new instance of the class",
                )(
                    lambda ast: ast.return_(
                        expr.ExprNewInstance(
                            expr.ExprIdent(cls.name), [PredefinedFn.dict(deser_args)]
                        )
                    )
                ),
                stmt.LineBreak(),
                lambda ast13: ast13.func(
                    "toDraft",
                    [],
                    expr.ExprIdent(f"Draft{cls.name}"),
                    comment="Convert the class instance to a draft for editing",
                )(
                    stmt.ReturnStatement(
                        expr.ExprMethodCall(
                            expr.ExprIdent(f"Draft{cls.name}"),
                            "update",
                            [expr.ExprIdent("this")],
                        )
                    ),
                ),
            ),
        )

        pkg.module(cls.name).write(program)

    def make_draft(cls: Class, pkg: Package):
        if not cls.is_public:
            # skip classes that are not public
            return

        idprop = cls.get_id_property()

        draft_clsname = "Draft" + cls.name
        draft_validators = f"draft{cls.name}Validators"

        program = Program()
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}", True)
        program.import_("mobx.makeObservable", True)
        program.import_("mobx.observable", True)
        program.import_("mobx.action", True)
        program.import_("sera-db.validators", True)

        import_helper = ImportHelper(program, TS_GLOBAL_IDENTS)

        program.root(
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                "const {getValidator, memoizeOneValidators} = validators;"
            ),
            stmt.LineBreak(),
        )

        # make sure that the property stale is not in existing properties
        if "stale" in cls.properties:
            raise ValueError(f"Class {cls.name} already has property stale")

        # information about class primary key
        cls_pk = None
        observable_args: list[tuple[expr.Expr, expr.ExprIdent]] = []
        prop_defs = []
        prop_validators: list[tuple[expr.ExprIdent, expr.Expr]] = []
        prop_constructor_assigns = []
        # attrs needed for the cls.create function
        create_args = []
        update_args = []
        ser_args = []
        to_record_args = []
        update_field_funcs: list[Callable[[AST], Any]] = []

        prop2tsname = {}

        for prop in cls.properties.values():
            # if prop.data.is_private:
            #     # skip private fields as this is for APIs exchange
            #     continue

            propname = to_camel_case(prop.name)
            prop2tsname[prop.name] = propname

            def _update_field_func(
                prop: DataProperty | ObjectProperty,
                propname: str,
                tstype: TsTypeWithDep,
                draft_clsname: str,
            ):
                return lambda ast: ast(
                    stmt.LineBreak(),
                    lambda ast01: ast01.func(
                        f"update{to_pascal_case(prop.name)}",
                        [
                            DeferredVar.simple(
                                "value",
                                expr.ExprIdent(tstype.type),
                            ),
                        ],
                        expr.ExprIdent(draft_clsname),
                        comment=f"Update the `{prop.name}` field",
                    )(
                        stmt.AssignStatement(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("this"), expr.ExprIdent(propname)
                            ),
                            expr.ExprIdent("value"),
                        ),
                        stmt.AssignStatement(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("this"), expr.ExprIdent("stale")
                            ),
                            expr.ExprConstant(True),
                        ),
                        stmt.ReturnStatement(expr.ExprIdent("this")),
                    ),
                )

            if isinstance(prop, DataProperty):
                tstype = prop.get_data_model_datatype().get_typescript_type()
                original_tstype = tstype

                if idprop is not None and prop.name == idprop.name:
                    # use id type alias
                    tstype = TsTypeWithDep(
                        f"{cls.name}Id",
                        deps=[f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}Id"],
                    )
                    original_tstype = tstype
                elif tstype.type not in schema.enums:
                    # for none id & none enum properties, we need to include a type for "invalid" value
                    tstype = _inject_type_for_invalid_value(tstype)

                if prop.is_optional:
                    # convert type to optional
                    tstype = tstype.as_optional_type()
                    original_tstype = original_tstype.as_optional_type()

                for dep in tstype.deps:
                    program.import_(dep, True)

                # however, if this is a primary key and auto-increment, we set a different default value
                # to be -1 to avoid start from 0
                if (
                    prop.db is not None
                    and prop.db.is_primary_key
                    and prop.db.is_auto_increment
                ):
                    create_propvalue = expr.ExprConstant(-1)
                elif is_set(prop.data.default_value):
                    create_propvalue = expr.ExprConstant(prop.data.default_value)
                else:
                    if tstype.type in idprop_aliases:
                        create_propvalue = idprop_aliases[tstype.type].get_default()
                    elif tstype.type in schema.enums:
                        create_propvalue = expr.ExprConstant(
                            next(iter(schema.enums[tstype.type].values.values())).value
                        )
                    else:
                        create_propvalue = tstype.get_default()

                prop_validators.append(
                    (
                        expr.ExprIdent(propname),
                        expr.ExprFuncCall(
                            expr.ExprIdent("getValidator"),
                            [
                                PredefinedFn.list(
                                    [
                                        expr.ExprConstant(
                                            constraint.get_typescript_constraint()
                                        )
                                        for constraint in prop.data.constraints
                                    ]
                                ),
                            ],
                        ),
                    )
                )

                if prop.db is not None and prop.db.is_primary_key:
                    # for checking if the primary key is from the database or default (create_propvalue)
                    cls_pk = (expr.ExprIdent(propname), create_propvalue)

                # if this field is private, we cannot get it from the normal record
                # we have to create a default value for it.
                if prop.data.is_private:
                    update_propvalue = create_propvalue
                else:
                    update_propvalue = PredefinedFn.attr_getter(
                        expr.ExprIdent("record"), expr.ExprIdent(propname)
                    )

                if original_tstype.type != tstype.type:
                    norm_func = get_norm_func(original_tstype, import_helper)
                else:
                    norm_func = identity

                ser_args.append(
                    (
                        expr.ExprIdent(prop.name),
                        (
                            expr.ExprTernary(
                                PredefinedFn.attr_getter(
                                    expr.ExprFuncCall(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent(draft_validators),
                                            expr.ExprIdent(propname),
                                        ),
                                        [
                                            PredefinedFn.attr_getter(
                                                expr.ExprIdent("this"),
                                                expr.ExprIdent(propname),
                                            )
                                        ],
                                    ),
                                    expr.ExprIdent("isValid"),
                                ),
                                norm_func(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent("this"), expr.ExprIdent(propname)
                                    )
                                ),
                                expr.ExprIdent("undefined"),
                            )
                            if prop.is_optional
                            else norm_func(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("this"), expr.ExprIdent(propname)
                                )
                            )
                        ),
                    )
                )

                if not prop.data.is_private:
                    # private property does not include in the public record
                    to_record_args.append(
                        (
                            expr.ExprIdent(propname),
                            (
                                expr.ExprTernary(
                                    PredefinedFn.attr_getter(
                                        expr.ExprFuncCall(
                                            PredefinedFn.attr_getter(
                                                expr.ExprIdent(draft_validators),
                                                expr.ExprIdent(propname),
                                            ),
                                            [
                                                PredefinedFn.attr_getter(
                                                    expr.ExprIdent("this"),
                                                    expr.ExprIdent(propname),
                                                )
                                            ],
                                        ),
                                        expr.ExprIdent("isValid"),
                                    ),
                                    norm_func(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent("this"),
                                            expr.ExprIdent(propname),
                                        )
                                    ),
                                    expr.ExprIdent("undefined"),
                                )
                                if prop.is_optional
                                else norm_func(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent("this"), expr.ExprIdent(propname)
                                    )
                                )
                            ),
                        )
                    )
                if not (prop.db is not None and prop.db.is_primary_key):
                    # skip observable for primary key as it is not needed
                    observable_args.append(
                        (
                            expr.ExprIdent(propname),
                            expr.ExprIdent("observable"),
                        )
                    )
                    observable_args.append(
                        (
                            expr.ExprIdent(f"update{to_pascal_case(prop.name)}"),
                            expr.ExprIdent("action"),
                        )
                    )
            else:
                assert isinstance(prop, ObjectProperty)
                if prop.target.db is not None:
                    # this class is stored in the database, we store the id instead
                    tstype = TsTypeWithDep(
                        f"{prop.target.name}Id",
                        [
                            f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.name}.{prop.target.name}Id"
                        ],
                    )
                    if prop.cardinality.is_star_to_many():
                        tstype = tstype.as_list_type()
                        create_propvalue = expr.ExprConstant([])
                    else:
                        if prop.is_optional:
                            # convert type to optional - for list type, we don't need to do this
                            # as we will use empty list as no value
                            tstype = tstype.as_optional_type()
                        # if target class has an auto-increment primary key, we set a different default value
                        # to be -1 to avoid start from 0
                        target_idprop = prop.target.get_id_property()
                        if (
                            target_idprop is not None
                            and target_idprop.db is not None
                            and target_idprop.db.is_primary_key
                            and target_idprop.db.is_auto_increment
                        ):
                            create_propvalue = expr.ExprConstant(-1)
                        else:
                            assert tstype.type in idprop_aliases
                            create_propvalue = idprop_aliases[tstype.type].get_default()

                    update_propvalue = PredefinedFn.attr_getter(
                        expr.ExprIdent("record"), expr.ExprIdent(propname)
                    )
                    ser_args.append(
                        (
                            expr.ExprIdent(prop.name),
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("this"), expr.ExprIdent(propname)
                            ),
                        )
                    )

                    if not prop.data.is_private:
                        # private property does not include in the public record
                        to_record_args.append(
                            (
                                expr.ExprIdent(propname),
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("this"), expr.ExprIdent(propname)
                                ),
                            )
                        )
                else:
                    # we are going to store the whole object
                    tstype = TsTypeWithDep(
                        f"Draft{prop.target.name}",
                        [
                            f"@.models.{prop.target.get_tsmodule_name()}.Draft{prop.target.name}.Draft{prop.target.name}"
                        ],
                    )
                    if prop.cardinality.is_star_to_many():
                        create_propvalue = expr.ExprConstant([])
                        update_propvalue = PredefinedFn.map_list(
                            PredefinedFn.attr_getter(
                                expr.ExprIdent("record"), expr.ExprIdent(propname)
                            ),
                            lambda item: expr.ExprMethodCall(
                                expr.ExprIdent(tstype.type),
                                "update",
                                [item],
                            ),
                        )
                        ser_args.append(
                            (
                                expr.ExprIdent(prop.name),
                                PredefinedFn.map_list(
                                    PredefinedFn.attr_getter(
                                        expr.ExprIdent("this"), expr.ExprIdent(propname)
                                    ),
                                    lambda item: expr.ExprMethodCall(item, "ser", []),
                                    (
                                        (
                                            lambda item: PredefinedFn.attr_getter(
                                                expr.ExprFuncCall(
                                                    PredefinedFn.attr_getter(
                                                        expr.ExprIdent(
                                                            draft_validators
                                                        ),
                                                        expr.ExprIdent(propname),
                                                    ),
                                                    [item],
                                                ),
                                                expr.ExprIdent("isValid"),
                                            )
                                        )
                                        if prop.is_optional
                                        else None
                                    ),
                                ),
                            )
                        )

                        if not prop.data.is_private:
                            # private property does not include in the public record
                            to_record_args.append(
                                (
                                    expr.ExprIdent(propname),
                                    PredefinedFn.map_list(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent("this"),
                                            expr.ExprIdent(propname),
                                        ),
                                        lambda item: expr.ExprMethodCall(
                                            item, "toRecord", []
                                        ),
                                        (
                                            (
                                                lambda item: PredefinedFn.attr_getter(
                                                    expr.ExprFuncCall(
                                                        PredefinedFn.attr_getter(
                                                            expr.ExprIdent(
                                                                draft_validators
                                                            ),
                                                            expr.ExprIdent(propname),
                                                        ),
                                                        [item],
                                                    ),
                                                    expr.ExprIdent("isValid"),
                                                )
                                            )
                                            if prop.is_optional
                                            else None
                                        ),
                                    ),
                                )
                            )

                        tstype = tstype.as_list_type()
                    else:
                        create_propvalue = expr.ExprMethodCall(
                            expr.ExprIdent(tstype.type),
                            "create",
                            [],
                        )
                        update_propvalue = expr.ExprMethodCall(
                            expr.ExprIdent(tstype.type),
                            "update",
                            [
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("record"), expr.ExprIdent(propname)
                                ),
                            ],
                        )

                        if prop.is_optional:
                            ser_args.append(
                                (
                                    expr.ExprIdent(prop.name),
                                    expr.ExprTernary(
                                        PredefinedFn.attr_getter(
                                            expr.ExprFuncCall(
                                                PredefinedFn.attr_getter(
                                                    expr.ExprIdent(draft_validators),
                                                    expr.ExprIdent(propname),
                                                ),
                                                [
                                                    PredefinedFn.attr_getter(
                                                        expr.ExprIdent("this"),
                                                        expr.ExprIdent(propname),
                                                    )
                                                ],
                                            ),
                                            expr.ExprIdent("isValid"),
                                        ),
                                        expr.ExprMethodCall(
                                            PredefinedFn.attr_getter(
                                                expr.ExprIdent("this"),
                                                expr.ExprIdent(propname),
                                            ),
                                            "ser",
                                            [],
                                        ),
                                        expr.ExprIdent("undefined"),
                                    ),
                                )
                            )
                            if not prop.data.is_private:
                                # private property does not include in the public record
                                to_record_args.append(
                                    (
                                        expr.ExprIdent(propname),
                                        expr.ExprTernary(
                                            PredefinedFn.attr_getter(
                                                expr.ExprFuncCall(
                                                    PredefinedFn.attr_getter(
                                                        expr.ExprIdent(
                                                            draft_validators
                                                        ),
                                                        expr.ExprIdent(propname),
                                                    ),
                                                    [
                                                        PredefinedFn.attr_getter(
                                                            expr.ExprIdent("this"),
                                                            expr.ExprIdent(propname),
                                                        )
                                                    ],
                                                ),
                                                expr.ExprIdent("isValid"),
                                            ),
                                            expr.ExprMethodCall(
                                                PredefinedFn.attr_getter(
                                                    expr.ExprIdent("this"),
                                                    expr.ExprIdent(propname),
                                                ),
                                                "toRecord",
                                                [],
                                            ),
                                            expr.ExprIdent("undefined"),
                                        ),
                                    )
                                )
                        else:
                            ser_args.append(
                                (
                                    expr.ExprIdent(prop.name),
                                    expr.ExprMethodCall(
                                        PredefinedFn.attr_getter(
                                            expr.ExprIdent("this"),
                                            expr.ExprIdent(propname),
                                        ),
                                        "ser",
                                        [],
                                    ),
                                )
                            )
                            if not prop.data.is_private:
                                # private property does not include in the public record
                                to_record_args.append(
                                    (
                                        expr.ExprIdent(propname),
                                        expr.ExprMethodCall(
                                            PredefinedFn.attr_getter(
                                                expr.ExprIdent("this"),
                                                expr.ExprIdent(propname),
                                            ),
                                            "toRecord",
                                            [],
                                        ),
                                    )
                                )

                        if prop.is_optional:
                            # convert type to optional - for list type, we don't need to do this
                            # as we will use empty list as no value
                            tstype = tstype.as_optional_type()

                for dep in tstype.deps:
                    program.import_(
                        dep,
                        True,
                    )

                observable_args.append(
                    (
                        expr.ExprIdent(propname),
                        expr.ExprIdent("observable"),
                    )
                )
                observable_args.append(
                    (
                        expr.ExprIdent(f"update{to_pascal_case(prop.name)}"),
                        expr.ExprIdent("action"),
                    )
                )

                # TODO: fix me!
                prop_validators.append(
                    (
                        expr.ExprIdent(propname),
                        expr.ExprFuncCall(
                            expr.ExprIdent("getValidator"),
                            [
                                PredefinedFn.list(
                                    [
                                        expr.ExprConstant(
                                            constraint.get_typescript_constraint()
                                        )
                                        for constraint in prop.data.constraints
                                    ]
                                ),
                            ],
                        ),
                    )
                )

            prop_defs.append(stmt.DefClassVarStatement(propname, tstype.type))
            prop_constructor_assigns.append(
                stmt.AssignStatement(
                    PredefinedFn.attr_getter(
                        expr.ExprIdent("this"),
                        expr.ExprIdent(propname),
                    ),
                    expr.ExprIdent("args." + propname),
                )
            )
            create_args.append((expr.ExprIdent(propname), create_propvalue))
            update_args.append(
                (
                    expr.ExprIdent(propname),
                    # if this is mutable property, we need to copy to make it immutable.
                    clone_prop(prop, update_propvalue),
                )
            )
            update_field_funcs.append(
                _update_field_func(prop, propname, tstype, draft_clsname)
            )

        prop_defs.append(stmt.DefClassVarStatement("stale", "boolean"))
        prop_constructor_assigns.append(
            stmt.AssignStatement(
                PredefinedFn.attr_getter(
                    expr.ExprIdent("this"), expr.ExprIdent("stale")
                ),
                expr.ExprIdent("args.stale"),
            )
        )
        observable_args.append(
            (
                expr.ExprIdent("stale"),
                expr.ExprIdent("observable"),
            )
        )
        create_args.append(
            (
                expr.ExprIdent("stale"),
                expr.ExprConstant(True),
            ),
        )
        update_args.append(
            (
                expr.ExprIdent("stale"),
                expr.ExprConstant(False),
            ),
        )
        observable_args.sort(key=lambda x: {"observable": 0, "action": 1}[x[1].ident])

        validators = expr.ExprFuncCall(
            expr.ExprIdent("memoizeOneValidators"), [PredefinedFn.dict(prop_validators)]
        )

        program.root(
            lambda ast00: ast00.class_like(
                "interface",
                draft_clsname + "ConstructorArgs",
            )(*prop_defs),
            stmt.LineBreak(),
            lambda ast10: ast10.class_(draft_clsname)(
                *prop_defs,
                stmt.LineBreak(),
                lambda ast10: ast10.func(
                    "constructor",
                    [
                        DeferredVar.simple(
                            "args",
                            expr.ExprIdent(draft_clsname + "ConstructorArgs"),
                        ),
                    ],
                )(
                    *prop_constructor_assigns,
                    stmt.LineBreak(),
                    stmt.SingleExprStatement(
                        expr.ExprFuncCall(
                            expr.ExprIdent("makeObservable"),
                            [
                                expr.ExprIdent("this"),
                                PredefinedFn.dict(observable_args),
                            ],
                        )
                    ),
                ),
                stmt.LineBreak(),
                lambda ast11: (
                    ast11.func(
                        "isNewRecord",
                        [],
                        expr.ExprIdent("boolean"),
                        comment="Check if this draft is for creating a new record",
                    )(
                        stmt.ReturnStatement(
                            expr.ExprEqual(
                                PredefinedFn.attr_getter(
                                    expr.ExprIdent("this"), cls_pk[0]
                                ),
                                cls_pk[1],
                            )
                        )
                    )
                    if cls_pk is not None
                    else None
                ),
                stmt.LineBreak(),
                lambda ast12: ast12.func(
                    "create",
                    [],
                    expr.ExprIdent(draft_clsname),
                    is_static=True,
                    comment="Make a new draft for creating a new record",
                )(
                    stmt.ReturnStatement(
                        expr.ExprNewInstance(
                            expr.ExprIdent(draft_clsname),
                            [PredefinedFn.dict(create_args)],
                        )
                    ),
                ),
                stmt.LineBreak(),
                lambda ast13: ast13.func(
                    "update",
                    [DeferredVar.simple("record", expr.ExprIdent(cls.name))],
                    expr.ExprIdent(draft_clsname),
                    is_static=True,
                    comment="Make a new draft for updating an existing record",
                )(
                    stmt.ReturnStatement(
                        expr.ExprNewInstance(
                            expr.ExprIdent(draft_clsname),
                            [PredefinedFn.dict(update_args)],
                        )
                    ),
                ),
                *update_field_funcs,
                stmt.LineBreak(),
                lambda ast14: ast14.func(
                    "isValid",
                    [],
                    expr.ExprIdent("boolean"),
                    comment="Check if the draft is valid (only check the required fields as the non-required fields if it's invalid will be set to undefined)",
                )(
                    stmt.ReturnStatement(
                        expr.ExprRawTypescript(
                            " && ".join(
                                f"{draft_validators}.{prop2tsname[prop.name]}(this.{prop2tsname[prop.name]}).isValid"
                                for prop in cls.properties.values()
                                if not prop.is_optional
                            )
                            if any(
                                not prop.is_optional for prop in cls.properties.values()
                            )
                            else "true"
                        )
                    )
                ),
                stmt.LineBreak(),
                lambda ast15: ast15.func(
                    "ser",
                    [],
                    expr.ExprIdent("any"),
                    comment="Serialize the draft to communicate with the server. `isValid` must be called first to ensure all data is valid",
                )(
                    stmt.ReturnStatement(
                        PredefinedFn.dict(ser_args),
                    ),
                ),
                stmt.LineBreak(),
                lambda ast16: ast16.func(
                    "toRecord",
                    [],
                    expr.ExprIdent(cls.name),
                    comment="Convert the draft to a normal record. `isValid` must be called first to ensure all data is valid",
                )(
                    stmt.ReturnStatement(
                        expr.ExprNewInstance(
                            expr.ExprIdent(cls.name),
                            [PredefinedFn.dict(to_record_args)],
                        ),
                    )
                ),
            ),
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                f"export const {draft_validators} = " + validators.to_typescript() + ";"
            ),
        )

        pkg.module("Draft" + cls.name).write(program)

    def make_table(cls: Class, pkg: Package):
        if not cls.is_public or cls.db is None:
            # skip classes that are not public and not stored in the database
            return

        outmod = pkg.module(cls.name + "Table")
        if outmod.exists():
            # skip if the module already exists
            logger.info(f"Module {outmod.path} already exists, skip")
            return

        program = Program()
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}", True)
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}Id", True)
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}Query.query", True)
        program.import_(
            f"@.models.{pkg.dir.name}.Draft{cls.name}.Draft{cls.name}", True
        )
        program.import_("sera-db.Table", True)
        program.import_("sera-db.DB", True)

        program.root(
            stmt.LineBreak(),
            lambda ast00: ast00.class_(
                f"{cls.name}Table",
                [expr.ExprIdent(f"Table<{cls.name}Id, {cls.name}, Draft{cls.name}>")],
            )(
                lambda ast01: ast01.func(
                    "constructor",
                    [
                        DeferredVar.simple(
                            "db",
                            expr.ExprIdent("DB"),
                        )
                    ],
                )(
                    stmt.SingleExprStatement(
                        expr.ExprFuncCall(
                            expr.ExprIdent("super"),
                            [
                                PredefinedFn.dict(
                                    [
                                        (
                                            expr.ExprIdent("cls"),
                                            expr.ExprIdent(cls.name),
                                        ),
                                        (
                                            expr.ExprIdent("remoteURL"),
                                            expr.ExprConstant(
                                                f"/api/{to_snake_case(cls.name).replace('_', '-')}"
                                            ),
                                        ),
                                        (
                                            expr.ExprIdent("db"),
                                            expr.ExprIdent("db"),
                                        ),
                                        (
                                            expr.ExprIdent("queryProcessor"),
                                            expr.ExprIdent("query"),
                                        ),
                                    ]
                                )
                            ],
                        )
                    )
                ),
            ),
        )

        outmod.write(program)

    def make_query_processor(cls: Class, pkg: Package):
        if not cls.is_public:
            # skip classes that are not public
            return

        outmod = pkg.module(cls.name + "Query")

        program = Program()
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}", True)
        program.import_(f"sera-db.QueryProcessor", True)

        query_args = []
        for prop in cls.properties.values():
            propname = to_camel_case(prop.name)
            if propname != prop.name:
                query_args.append(
                    (
                        expr.ExprIdent(propname),
                        expr.ExprConstant(prop.name),
                    )
                )

        program.root(
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                f"export const query = "
                + expr.ExprNewInstance(
                    expr.ExprIdent(f"QueryProcessor<{cls.name}>"),
                    [
                        PredefinedFn.dict(query_args),
                    ],
                ).to_typescript()
                + ";",
            ),
        )

        outmod.write(program)

    def make_definition(cls: Class, pkg: Package):
        """Make schema definition for the class in frontend so that we can generate components"""
        if not cls.is_public:
            # skip classes that are not public
            return

        program = Program()
        prop_defs: list[tuple[DataProperty | ObjectProperty, expr.Expr, expr.Expr]] = []
        prop_normalizers: list[tuple[expr.Expr, expr.Expr]] = []

        import_helper = ImportHelper(program, TS_GLOBAL_IDENTS)

        for prop in cls.properties.values():
            # we must include private properties that are needed during upsert for our forms.
            # if prop.data.is_private:
            #     # skip private fields as this is for APIs exchange
            #     continue
            propname = to_camel_case(prop.name)
            tsprop = {}

            if isinstance(prop, DataProperty):
                tstype = prop.get_data_model_datatype().get_typescript_type()
                # for schema definition, we need to use the original type, not the type alias
                # if prop.name == idprop.name:
                #     # use id type alias
                #     tstype = TsTypeWithDep(f"{cls.name}Id")
                for dep in tstype.deps:
                    program.import_(dep, True)
                tsprop = [
                    (
                        expr.ExprIdent("datatype"),
                        (
                            expr.ExprConstant(tstype.type)
                            if tstype.type not in schema.enums
                            else expr.ExprConstant("enum")
                        ),
                    ),
                    *(
                        [(expr.ExprIdent("enumType"), expr.ExprIdent(tstype.type))]
                        if tstype.type in schema.enums
                        else []
                    ),
                    (
                        expr.ExprIdent("isRequired"),
                        expr.ExprConstant(
                            not prop.is_optional
                            and prop.default_value is None
                            and prop.default_factory is None
                        ),
                    ),
                ]

                norm_func = get_normalizer(tstype, import_helper)
                if norm_func is not None:
                    # we have a normalizer for this type
                    prop_normalizers.append((expr.ExprIdent(propname), norm_func))
            else:
                assert isinstance(prop, ObjectProperty)
                if prop.target.db is not None:
                    # this class is stored in the database, we store the id instead
                    tstype = (
                        assert_not_null(prop.target.get_id_property())
                        .get_data_model_datatype()
                        .get_typescript_type()
                    )
                else:
                    # we are going to store the whole object
                    tstype = TsTypeWithDep(
                        prop.target.name,
                        [
                            f"@.models.{prop.target.get_tsmodule_name()}.{prop.target.name}.{prop.target.name}"
                        ],
                    )

                # we don't store the type itself, but just the name of the type
                # so not need to import the dependency
                # if tstype.dep is not None:
                #     program.import_(
                #         tstype.dep,
                #         True,
                #     )

                tsprop = [
                    (
                        expr.ExprIdent("targetClass"),
                        expr.ExprConstant(prop.target.name),
                    ),
                    (
                        expr.ExprIdent("datatype"),
                        expr.ExprConstant(
                            tstype.type if prop.target.db is not None else "undefined"
                        ),
                    ),
                    (
                        expr.ExprIdent("cardinality"),
                        expr.ExprConstant(prop.cardinality.value),
                    ),
                    (
                        expr.ExprIdent("isEmbedded"),
                        expr.ExprConstant(prop.target.db is None),
                    ),
                    (
                        expr.ExprIdent("isRequired"),
                        expr.ExprConstant(not prop.is_optional),
                    ),
                ]

            prop_defs.append(
                (
                    prop,
                    expr.ExprIdent(propname),
                    PredefinedFn.dict(
                        [
                            (expr.ExprIdent("name"), expr.ExprConstant(prop.name)),
                            (expr.ExprIdent("tsName"), expr.ExprConstant(propname)),
                            (
                                expr.ExprIdent("updateFuncName"),
                                expr.ExprConstant(f"update{to_pascal_case(prop.name)}"),
                            ),
                            (
                                expr.ExprIdent("label"),
                                expr.ExprConstant(prop.label.to_dict()),
                            ),
                            (
                                expr.ExprIdent("description"),
                                (
                                    expr.ExprConstant(prop.description.to_dict())
                                    if not prop.description.is_empty()
                                    else expr.ExprConstant("undefined")
                                ),
                            ),
                            (
                                expr.ExprIdent("constraints"),
                                PredefinedFn.list(
                                    [
                                        expr.ExprConstant(
                                            constraint.get_typescript_constraint()
                                        )
                                        for constraint in prop.data.constraints
                                    ]
                                ),
                            ),
                        ]
                        + tsprop
                    ),
                )
            )

        for type in ["ObjectProperty", "DataProperty"]:
            program.import_(f"sera-db.{type}", True)
        if cls.db is not None:
            program.import_(f"sera-db.Schema", True)
        else:
            program.import_(f"sera-db.EmbeddedSchema", True)

        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}", True)
        program.import_(
            f"@.models.{pkg.dir.name}.Draft{cls.name}.Draft{cls.name}", True
        )
        program.import_(
            f"@.models.{pkg.dir.name}.Draft{cls.name}.draft{cls.name}Validators", True
        )
        if cls.db is not None:
            program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}Id", True)

        program.root(
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                f"export type {cls.name}SchemaType = "
                + PredefinedFn.dict(
                    (
                        [
                            (expr.ExprIdent("id"), expr.ExprIdent(f"{cls.name}Id")),
                        ]
                        if cls.db is not None
                        else []
                    )
                    + [
                        (
                            expr.ExprIdent("publicProperties"),
                            expr.ExprIdent(
                                " | ".join(
                                    [
                                        expr.ExprConstant(
                                            to_camel_case(prop.name)
                                        ).to_typescript()
                                        for prop in cls.properties.values()
                                        if not prop.data.is_private
                                    ]
                                )
                            ),
                        ),
                        (
                            expr.ExprIdent("allProperties"),
                            expr.ExprIdent(
                                f"{cls.name}SchemaType['publicProperties']"
                                + (
                                    " | "
                                    + " | ".join(
                                        [
                                            expr.ExprConstant(
                                                to_camel_case(prop.name)
                                            ).to_typescript()
                                            for prop in cls.properties.values()
                                            if prop.data.is_private
                                        ]
                                    )
                                    if any(
                                        prop.data.is_private
                                        for prop in cls.properties.values()
                                    )
                                    else ""
                                )
                            ),
                        ),
                        (
                            expr.ExprIdent("cls"),
                            expr.ExprIdent(cls.name),
                        ),
                        (
                            expr.ExprIdent("draftCls"),
                            expr.ExprIdent(f"Draft{cls.name}"),
                        ),
                    ]
                ).to_typescript()
                + ";",
            ),
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                f"const publicProperties: Record<{cls.name}SchemaType['publicProperties'], DataProperty | ObjectProperty> = "
                + PredefinedFn.dict(
                    [
                        (prop_name, prop_def)
                        for prop, prop_name, prop_def in prop_defs
                        if not prop.data.is_private
                    ]
                ).to_typescript()
                + ";"
            ),
            stmt.LineBreak(),
            stmt.TypescriptStatement(
                (
                    f"export const {cls.name}Schema: Schema<{cls.name}SchemaType['id'], {cls.name}SchemaType['cls'], {cls.name}SchemaType['draftCls'], {cls.name}SchemaType['publicProperties'], {cls.name}SchemaType['allProperties'], {cls.name}SchemaType> = "
                    if cls.db is not None
                    else f"export const {cls.name}Schema: EmbeddedSchema<{cls.name}SchemaType['cls'], {cls.name}SchemaType['draftCls'], {cls.name}SchemaType['publicProperties'], {cls.name}SchemaType['allProperties']> = "
                )
                + PredefinedFn.dict(
                    [
                        (
                            expr.ExprIdent("publicProperties"),
                            expr.ExprIdent("publicProperties"),
                        ),
                        (
                            expr.ExprIdent("allProperties"),
                            expr.ExprIdent(
                                "{ ...publicProperties, "
                                + ", ".join(
                                    [
                                        f"{prop_name.to_typescript()}: {prop_def.to_typescript()}"
                                        for prop, prop_name, prop_def in prop_defs
                                        if prop.data.is_private
                                    ]
                                )
                                + "}"
                            ),
                        ),
                        (
                            expr.ExprIdent("validators"),
                            expr.ExprIdent(f"draft{cls.name}Validators"),
                        ),
                        (
                            expr.ExprIdent("normalizers"),
                            PredefinedFn.dict(prop_normalizers),
                        ),
                    ]
                    + (
                        [
                            (
                                expr.ExprIdent("primaryKey"),
                                expr.ExprConstant(
                                    assert_not_null(cls.get_id_property()).name
                                ),
                            )
                        ]
                        if cls.db is not None
                        else []
                    )
                ).to_typescript()
                + ";"
            ),
        )
        pkg.module(cls.name + "Schema").write(program)

    def make_index(pkg: Package):
        outmod = pkg.module("index")
        if outmod.exists():
            # skip if the module already exists
            logger.info(f"Module {outmod.path} already exists, skip")
            return

        export_types = []
        export_iso_types = []  # isolatedModules required separate export type clause

        program = Program()
        program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}", True)
        export_types.append(cls.name)
        if cls.db is not None:
            # only import the id if this class is stored in the database
            program.import_(f"@.models.{pkg.dir.name}.{cls.name}.{cls.name}Id", True)
            export_iso_types.append(f"{cls.name}Id")

        program.import_(
            f"@.models.{pkg.dir.name}.{cls.name}Schema.{cls.name}Schema", True
        )
        export_types.append(f"{cls.name}Schema")
        program.import_(
            f"@.models.{pkg.dir.name}.{cls.name}Schema.{cls.name}SchemaType", True
        )
        export_iso_types.append(f"{cls.name}SchemaType")

        program.import_(
            f"@.models.{pkg.dir.name}.Draft{cls.name}.Draft{cls.name}", True
        )
        export_types.append(f"Draft{cls.name}")
        if cls.db is not None:
            program.import_(
                f"@.models.{pkg.dir.name}.{cls.name}Table.{cls.name}Table", True
            )
            export_types.append(f"{cls.name}Table")

        program.root(
            stmt.LineBreak(),
            stmt.TypescriptStatement("export { %s };" % (", ".join(export_types))),
            (
                stmt.TypescriptStatement(
                    "export type { %s };" % (", ".join(export_iso_types))
                )
            ),
        )

        outmod.write(program)

    for cls in schema.topological_sort():
        pkg = target_pkg.pkg(cls.get_tsmodule_name())
        make_normal(cls, pkg)
        make_draft(cls, pkg)
        make_query_processor(cls, pkg)
        make_table(cls, pkg)
        make_definition(cls, pkg)

        make_index(pkg)


def make_typescript_enum(schema: Schema, target_pkg: Package):
    """Make typescript enum for the schema"""
    enum_pkg = target_pkg.pkg("enums")

    def make_enum(enum: Enum, pkg: Package):
        program = Program()
        program.root(
            stmt.LineBreak(),
            lambda ast: ast.class_like("enum", enum.name)(
                *[
                    stmt.DefEnumValueStatement(
                        name=value.name,
                        value=expr.ExprConstant(value.value),
                    )
                    for value in enum.values.values()
                ]
            ),
        )
        pkg.module(enum.get_tsmodule_name()).write(program)

    for enum in schema.enums.values():
        make_enum(enum, enum_pkg)

    program = Program()
    for enum in schema.enums.values():
        program.import_(f"@.models.enums.{enum.get_tsmodule_name()}.{enum.name}", True)

    program.root(
        stmt.LineBreak(),
        stmt.TypescriptStatement(
            "export { "
            + ", ".join([enum.name for enum in schema.enums.values()])
            + "};"
        ),
    )
    enum_pkg.module("index").write(program)


def _inject_type_for_invalid_value(tstype: TsTypeWithDep) -> TsTypeWithDep:
    """
    Inject a type for "invalid" values into the given TypeScript type. For context, see the discussion in Data Modeling Problems:
    What would be an appropriate type for an invalid value? Since it's user input, it will be a string type.

    However, there are some exceptions such as boolean type, which will always be valid and do not need injection.

    If the type already includes `string` type, no changes are needed. Otherwise, we add `string` to the type. For example:
    - (number | undefined) -> (number | undefined | string)
    - number | undefined -> number | undefined | string
    - number[] -> (number | string)[]
    - (number | undefined)[] -> (number | undefined | string)[]
    """
    if tstype.type == "boolean":
        return tstype

    # TODO: fix me and make it more robust!
    m = re.match(r"(\(?[a-zA-Z \|]+\)?)(\[\])", tstype.type)
    if m is not None:
        # This is an array type, add string to the inner type
        inner_type = m.group(1)
        if "string" not in inner_type:
            if inner_type.startswith("(") and inner_type.endswith(")"):
                # Already has parentheses
                inner_type = f"{inner_type[:-1]} | string)"
            else:
                # Need to add parentheses
                inner_type = f"({inner_type} | string)"
        return TsTypeWithDep(inner_type + "[]", tstype.deps)

    m = re.match(r"^\(?[a-zA-Z \|]+\)?$", tstype.type)
    if m is not None:
        if "string" not in tstype.type:
            if tstype.type.startswith("(") and tstype.type.endswith(")"):
                # Already has parentheses
                new_type = f"{tstype.type[:-1]} | string)"
            else:
                # Needs parentheses for clarity
                new_type = f"({tstype.type} | string)"
            return TsTypeWithDep(new_type, tstype.deps)
        return tstype

    raise NotImplementedError(tstype.type)


def get_normalizer(
    tstype: TsTypeWithDep, import_helper: ImportHelper
) -> Optional[expr.ExprIdent]:
    if tstype.type == "number":
        return import_helper.use("normalizers.normalizeNumber")
    if tstype.type == "number | undefined":
        return import_helper.use("normalizers.normalizeOptionalNumber")

    assert "number" not in tstype.type, tstype.type
    return None


def get_norm_func(
    tstype: TsTypeWithDep, import_helper: ImportHelper
) -> Callable[[expr.Expr], expr.Expr]:
    """
    Get the normalizer function for the given TypeScript type.
    If no normalizer is available, return None.
    """
    norm_func = get_normalizer(tstype, import_helper)
    if norm_func is not None:

        def modify_expr(value: expr.Expr) -> expr.Expr:
            return expr.ExprFuncCall(
                norm_func,
                [value],
            )

        return modify_expr
    return identity  # Return the value as is if no normalizer is available



from __future__ import annotations

from typing import Annotated, Any, Callable, Optional

from graph.interface import BaseEdge

from sera.libs.directed_computing_graph._node import NodeId
from sera.libs.directed_computing_graph._type_conversion import UnitTypeConversion


class DCGEdge(BaseEdge[NodeId, int]):

    def __init__(
        self,
        id: int,
        source: NodeId,
        target: NodeId,
        argindex: int,
        type_conversion: UnitTypeConversion,
        filter_fn: Optional[Callable[[Any], bool]] = None,
    ):
        super().__init__(id, source, target, key=argindex)
        self.argindex = argindex
        self.type_conversion = type_conversion
        self.filter_fn = filter_fn

    def filter(self, value: Any) -> bool:
        """Filter the value passing through this edge.

        Returns:
            True if the value should flow through this edge, False to block it.
        """
        if self.filter_fn is not None:
            return self.filter_fn(value)
        return True



from __future__ import annotations

import asyncio
import inspect
from dataclasses import dataclass
from enum import Enum
from typing import (
    Annotated,
    Any,
    Awaitable,
    Callable,
    MutableSequence,
    Optional,
    Sequence,
)

from graph.retworkx import RetworkXStrDiGraph

from sera.libs.directed_computing_graph._edge import DCGEdge
from sera.libs.directed_computing_graph._flow import Flow
from sera.libs.directed_computing_graph._node import (
    ComputeFn,
    ComputeFnId,
    DCGNode,
    NodeId,
)
from sera.libs.directed_computing_graph._runtime import SKIP, NodeRuntime
from sera.libs.directed_computing_graph._type_conversion import (
    ComposeTypeConversion,
    TypeConversion,
    UnitTypeConversion,
    align_generic_type,
    ground_generic_type,
    is_generic_type,
)
from sera.misc import identity

TaskKey = Annotated[tuple, "TaskKey"]
TaskArgs = Annotated[MutableSequence, "TaskArgs"]


class DirectedComputingGraph:
    """
    A Directed Computing Graph (DCG) is a directed graph where nodes represent functions
    and edges represent dependencies between these functions. The graph is used to manage
    the execution of functions in a specific order based on their dependencies.
    """

    def __init__(
        self,
        graph: RetworkXStrDiGraph[int, DCGNode, DCGEdge],
        type_service: TypeConversion,
    ):
        self.graph = graph
        self.type_service = type_service

    @staticmethod
    def from_flows(
        flows: dict[ComputeFnId, Flow | ComputeFn],
        type_conversions: Optional[
            Sequence[UnitTypeConversion | ComposeTypeConversion]
        ] = None,
        strict: bool = True,
    ):
        """Create a computing graph from flow mapping.

        Args:
            flows: A dictionary mapping identifier to:
                1. a function
                2. a flow specifying the upstream functions and the function.
            type_conversions: A list of type conversions to be used for converting the input types.
            strict: If True, we do type checking.
        Returns:
            DirectedComputingGraph: A directed computing graph constructed from the provided flows.
        """
        # add typing conversions
        upd_type_conversions: list[UnitTypeConversion | ComposeTypeConversion] = list(
            type_conversions or []
        )
        type_service = TypeConversion(upd_type_conversions)

        g: RetworkXStrDiGraph[int, DCGNode, DCGEdge] = RetworkXStrDiGraph(
            check_cycle=False, multigraph=False
        )

        # create a graph
        for uid, uinfo in flows.items():
            if isinstance(uinfo, Flow):
                actor = uinfo.target
            else:
                actor = uinfo
            g.add_node(DCGNode(uid, actor))

        # create a graph
        for uid, uinfo in flows.items():
            if isinstance(uinfo, Flow):
                func = uinfo.target
            else:
                func = uinfo
            g.add_node(DCGNode(uid, func))

        # grounding function that has generic type input and output
        for uid, flow in flows.items():
            if not isinstance(flow, Flow):
                continue

            u = g.get_node(uid)
            usig = u.signature
            if is_generic_type(usig.return_type) or any(
                is_generic_type(t) for t in usig.argtypes
            ):
                var2type = {}
                for i, t in enumerate(usig.argtypes):
                    if is_generic_type(t):
                        # align the generic type with the previous return type
                        if len(flow.source) <= i and strict:
                            raise TypeConversion.UnknownConversion(
                                f"Cannot ground the generic type based on upstream actors for actor {uid}"
                            )

                        source_return_type = g.get_node(
                            flow.source[i]
                        ).signature.return_type

                        try:
                            usig.argtypes[i], (var, nt) = align_generic_type(
                                t, source_return_type
                            )
                        except Exception as e:
                            raise TypeConversion.UnknownConversion(
                                f"Cannot align the generic type {t} based on upstream actors for actor {uid}"
                            )
                        var2type[var] = nt
                if is_generic_type(usig.return_type):
                    usig.return_type = ground_generic_type(
                        usig.return_type,
                        var2type,
                    )

        for uid, flow in flows.items():
            if not isinstance(flow, Flow):
                continue

            u = g.get_node(uid)
            usig = u.signature
            for idx, sid in enumerate(flow.source):
                s = g.get_node(sid)
                ssig = s.signature
                cast_fn = identity
                try:
                    cast_fn = type_service.get_conversion(
                        ssig.return_type, usig.argtypes[idx]
                    )
                except Exception as e:
                    if strict:
                        raise TypeConversion.UnknownConversion(
                            f"Don't know how to convert output of `{sid}` to input of `{uid}`"
                        ) from e
                g.add_edge(
                    DCGEdge(
                        id=-1,
                        source=sid,
                        target=uid,
                        argindex=idx,
                        filter_fn=flow.filter_fn,
                        type_conversion=cast_fn,
                    )
                )

        # postprocessing such as type conversion, and args/context
        for u in g.iter_nodes():
            inedges = g.in_edges(u.id)

            # update the type conversion
            u.type_conversions = [identity] * len(u.signature.argnames)
            for inedge in inedges:
                u.type_conversions[inedge.argindex] = inedge.type_conversion

            # update the required args and context
            u.required_args = u.signature.argnames[: g.in_degree(u.id)]
            # arguments of a compute function that are not provided by the upstream actors must be provided by the context.
            u.required_context = u.signature.argnames[g.in_degree(u.id) :]
            u.required_context_default_args = {
                k: u.signature.default_args[k]
                for k in u.required_context
                if k in u.signature.default_args
            }

        return DirectedComputingGraph(g, type_service)

    def execute(
        self,
        input: dict[ComputeFnId, tuple],
        output: set[str],
        context: Optional[
            dict[str, Callable | Any] | Callable[[], dict[str, Any]]
        ] = None,
    ):
        """
        Execute the directed computing graph with the given input and output specifications.

        Args:
            input: A dictionary mapping function identifiers to their input arguments.
            output: A set of function identifiers that should be executed.
            context: An optional context that can be a dictionary of functions or a single function.
        """
        assert all(
            isinstance(v, tuple) for v in input.values()
        ), "Input must be a tuple"

        if context is None:
            context = {}
        elif isinstance(context, Callable):
            context = context()
        else:
            context = {k: v() if callable(v) else v for k, v in context.items()}

        # This is a quick reactive algorithm, we may be able to do it better.
        # The idea is when all inputs of a function is available, we can execute a function.
        # We assume that the memory is large enough to hold all the functions and their inputs
        # in the memory.

        # we execute the computing nodes
        # when it's finished, we put the outgoing edges into a stack.
        runtimes: dict[NodeId, NodeRuntime] = {}

        for u in self.graph.iter_nodes():
            if u.id in input:
                # user provided input should supersede the context
                n_provided_args = len(input[u.id])
                n_consumed_context = n_provided_args - len(u.required_args)
            else:
                n_consumed_context = 0

            node_context = tuple(
                (
                    context[name]
                    if name in context
                    else u.required_context_default_args[name]
                )
                for name in u.required_context[n_consumed_context:]
            )

            runtimes[u.id] = NodeRuntime.from_node(self.graph, u, node_context)
        stack: list[NodeId] = []

        for id, args in input.items():
            runtimes[id].add_task((0,), list(args))
            stack.append(id)

        return_output = {id: [] for id in output}

        while len(stack) > 0:
            # pop the one from the stack and execute it.
            id = stack.pop()
            runtime = runtimes[id]

            # if there is enough data for the node, we can execute it.
            # if it is not, we just skip it and it will be added back to the stack by one of its parents.
            # so we don't miss it.
            if not runtime.has_enough_data():
                continue

            outedges = self.graph.out_edges(id)
            successors: Sequence[tuple[DCGEdge, DCGNode]] = [
                (edge, self.graph.get_node(edge.target)) for edge in outedges
            ]

            # run the tasks and pass the output to the successors
            for task_id, task in runtime.tasks.items():
                if any(arg is SKIP for arg in task):
                    task_output = SKIP
                else:
                    task_output = runtime.execute(task)

                for outedge, succ in successors:
                    runtimes[succ.id].add_task_args(
                        task_id,
                        id,
                        (
                            SKIP
                            if task_output is SKIP or not outedge.filter(task_output)
                            else task_output
                        ),
                    )

                if id in output and task_output is not SKIP:
                    return_output[id].append(task_output)

            # retrieve the outgoing nodes and push them into the stack
            for outedge, succ in successors:
                stack.append(succ.id)

        return return_output

    async def execute_async(
        self,
        input: dict[ComputeFnId, tuple],
        output: set[str],
        context: Optional[
            dict[str, Callable | Any] | Callable[[], dict[str, Any]]
        ] = None,
    ):
        """
        Asynchronously execute the directed computing graph with the given input and output specifications.
        This method handles both synchronous and asynchronous functions.

        Args:
            input: A dictionary mapping function identifiers to their input arguments.
            output: A set of function identifiers that should be executed.
            context: An optional context that can be a dictionary of functions or a single function.
        """
        assert all(
            isinstance(v, tuple) for v in input.values()
        ), "Input must be a tuple"

        if context is None:
            context = {}
        elif isinstance(context, Callable):
            context = context()
        else:
            context = {k: v() if callable(v) else v for k, v in context.items()}

        # This is a quick reactive algorithm, we may be able to do it better.
        # The idea is when all inputs of a function is available, we can execute a function.
        # We assume that the memory is large enough to hold all the functions and their inputs
        # in the memory.

        # we execute the computing nodes
        # when it's finished, we put the outgoing edges into a stack.
        runtimes: dict[NodeId, NodeRuntime] = {}

        for u in self.graph.iter_nodes():
            if u.id in input:
                # user provided input should supersede the context
                n_provided_args = len(input[u.id])
                n_consumed_context = n_provided_args - len(u.required_args)
            else:
                n_consumed_context = 0

            node_context = tuple(
                (
                    context[name]
                    if name in context
                    else u.required_context_default_args[name]
                )
                for name in u.required_context[n_consumed_context:]
            )

            runtimes[u.id] = NodeRuntime.from_node(self.graph, u, node_context)
        stack: list[NodeId] = []

        for id, args in input.items():
            runtimes[id].add_task((0,), list(args))
            stack.append(id)

        return_output = {id: [] for id in output}

        while len(stack) > 0:
            # pop the one from the stack and execute it.
            id = stack.pop()
            runtime = runtimes[id]

            # if there is enough data for the node, we can execute it.
            # if it is not, we just skip it and it will be added back to the stack by one of its parents.
            # so we don't miss it.
            if not runtime.has_enough_data():
                continue

            outedges = self.graph.out_edges(id)
            successors: Sequence[tuple[DCGEdge, DCGNode]] = [
                (edge, self.graph.get_node(edge.target)) for edge in outedges
            ]

            # run the tasks and pass the output to the successors
            for task_id, task in runtime.tasks.items():
                if any(arg is SKIP for arg in task):
                    task_output = SKIP
                else:
                    if runtime.node.signature.is_async:
                        task_output = await runtime.execute(task)
                    else:
                        task_output = runtime.execute(task)

                for outedge, succ in successors:
                    runtimes[succ.id].add_task_args(
                        task_id,
                        id,
                        (
                            SKIP
                            if task_output is SKIP or not outedge.filter(task_output)
                            else task_output
                        ),
                    )

                if id in output and task_output is not SKIP:
                    return_output[id].append(task_output)

            # retrieve the outgoing nodes and push them into the stack
            for outedge, succ in successors:
                stack.append(succ.id)

        return return_output



from __future__ import annotations

from enum import Enum
from math import dist
from typing import Annotated, Any, Generic, NamedTuple, Optional, Sequence, TypeVar

from litestar.exceptions import HTTPException
from sqlalchemy import Result, Select, delete, exists, func, select
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import load_only

from sera.libs.base_orm import BaseORM
from sera.misc import assert_not_null
from sera.models import Class
from sera.typing import FieldName, T, doc


class QueryOp(str, Enum):
    lt = "<"
    lte = "<="
    gt = ">"
    gte = ">="
    eq = "="
    ne = "!="
    # select records where values are in the given list
    in_ = "in"
    not_in = "not in"
    # for full text search
    fuzzy = "fuzzy"


Query = Annotated[
    dict[FieldName, dict[QueryOp, Annotated[Any, doc("query value")]]],
    doc("query operations"),
]
R = TypeVar("R", bound=BaseORM)
ID = TypeVar("ID")  # ID of a class
SqlResult = TypeVar("SqlResult", bound=Result)


class QueryResult(NamedTuple, Generic[R]):
    records: Sequence[R]
    total: int


class BaseAsyncService(Generic[ID, R]):

    instance = None

    def __init__(self, cls: Class, orm_cls: type[R]):
        self.cls = cls
        self.orm_cls = orm_cls
        self.id_prop = assert_not_null(cls.get_id_property())

        self._cls_id_prop = getattr(self.orm_cls, self.id_prop.name)
        self.is_id_auto_increment = assert_not_null(self.id_prop.db).is_auto_increment

    @classmethod
    def get_instance(cls):
        """Get the singleton instance of the service."""
        if cls.instance is None:
            # assume that the subclass overrides the __init__ method
            # so that we don't need to pass the class and orm_cls
            cls.instance = cls()  # type: ignore[call-arg]
        return cls.instance

    async def get(
        self,
        query: Query,
        limit: int,
        offset: int,
        unique: bool,
        sorted_by: list[str],
        group_by: list[str],
        fields: list[str],
        session: AsyncSession,
    ) -> QueryResult[R]:
        """Retrieving records matched a query.

        Args:
            query: The query to filter the records
            limit: The maximum number of records to return
            offset: The number of records to skip before returning results
            unique: Whether to return unique results only
            sorted_by: list of field names to sort by, prefix a field with '-' to sort that field in descending order
            group_by: list of field names to group by
            fields: list of field names to include in the results -- empty means all fields
        """
        q = self._select()
        if fields:
            q = q.options(
                load_only(*[getattr(self.orm_cls, field) for field in fields])
            )
        if unique:
            q = q.distinct()
        if sorted_by:
            for field in sorted_by:
                if field.startswith("-"):
                    q = q.order_by(getattr(self.orm_cls, field[1:]).desc())
                else:
                    q = q.order_by(getattr(self.orm_cls, field))
        if group_by:
            for field in group_by:
                q = q.group_by(getattr(self.orm_cls, field))

        cq = select(func.count()).select_from(q.subquery())
        rq = q.limit(limit).offset(offset)
        records = self._process_result(await session.execute(rq)).scalars().all()
        total = (await session.execute(cq)).scalar_one()
        return QueryResult(records, total)

    async def get_by_id(self, id: ID, session: AsyncSession) -> Optional[R]:
        """Retrieving a record by ID."""
        q = self._select().where(self._cls_id_prop == id)
        result = self._process_result(await session.execute(q)).scalar_one_or_none()
        return result

    async def has_id(self, id: ID, session: AsyncSession) -> bool:
        """Check whether we have a record with the given ID."""
        q = exists().where(self._cls_id_prop == id).select()
        result = (await session.execute(q)).scalar()
        return bool(result)

    async def create(self, record: R, session: AsyncSession) -> R:
        """Create a new record."""
        if self.is_id_auto_increment:
            setattr(record, self.id_prop.name, None)

        try:
            session.add(record)
            await session.flush()
        except IntegrityError:
            raise HTTPException(detail="Invalid request", status_code=409)
        return record

    async def update(self, record: R, session: AsyncSession) -> R:
        """Update an existing record."""
        await session.execute(record.get_update_query())
        return record

    def _select(self) -> Select:
        """Get the select statement for the class."""
        return select(self.orm_cls)

    def _process_result(self, result: SqlResult) -> SqlResult:
        """Process the result of a query."""
        return result

    async def truncate(self, session: AsyncSession) -> None:
        """Truncate the table."""
        await session.execute(delete(self.orm_cls))



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Annotated, Any, MutableSequence, Sequence

from graph.retworkx import RetworkXStrDiGraph

from sera.libs.directed_computing_graph._edge import DCGEdge
from sera.libs.directed_computing_graph._node import DCGNode, NodeId

TaskKey = Annotated[tuple, "TaskKey"]
TaskArgs = Annotated[MutableSequence, "TaskArgs"]


class ArgValueType(Enum):
    UNSET = "UNSET"
    SKIP = "SKIP"


UNSET = ArgValueType.UNSET
SKIP = ArgValueType.SKIP


@dataclass
class NodeRuntime:
    id: NodeId
    tasks: dict[TaskKey, TaskArgs]
    context: Sequence[Any]

    graph: RetworkXStrDiGraph[int, DCGNode, DCGEdge]
    node: DCGNode
    indegree: int
    # This is a mapping from parent node id to the index of the argument in the task.
    parent2argindex: dict[str, int]

    @staticmethod
    def from_node(
        graph: RetworkXStrDiGraph[int, DCGNode, DCGEdge],
        node: DCGNode,
        context: Sequence[Any],
    ) -> NodeRuntime:
        return NodeRuntime(
            id=node.id,
            tasks={},
            context=context,
            graph=graph,
            node=node,
            indegree=graph.in_degree(node.id),
            parent2argindex={
                edge.source: i
                # Map parent node ID to argument index based on sorted in-edge order
                for i, edge in enumerate(
                    sorted(graph.in_edges(node.id), key=lambda e: e.id)
                )
            },
        )

    def add_task(self, key: TaskKey, args: TaskArgs) -> NodeRuntime:
        """
        Add a task to the node runtime.

        Args:
            key: The key identifying the task.
            args: The arguments for the task.
        Returns:
            NodeRuntime: The updated node runtime with the new task added.
        """
        self.tasks[key] = args
        return self

    def add_task_args(
        self, key: TaskKey, parent_node: NodeId, argvalue: Any
    ) -> NodeRuntime:
        """
        Add an argument to an existing task.

        Args:
            key: The key identifying the task.
            parent_node: Identifier of the parent node from which the argument is coming.
            argvalue: The value of the argument to add.
        Returns:
            NodeRuntime: The updated node runtime with the new argument added to the task.
        """
        if key not in self.tasks:
            self.tasks[key] = [UNSET] * self.indegree
        self.tasks[key][self.parent2argindex[parent_node]] = argvalue
        return self

    def has_enough_data(self) -> bool:
        """
        Check if the node has enough data to execute its tasks.

        Returns:
            bool: True if the node has enough data, False otherwise.
        """
        return all(
            all(arg is not UNSET for arg in args) for args in self.tasks.values()
        )

    def execute(self, task: TaskArgs) -> Any:
        """
        Execute a task with the given context.

        Args:
            task (TaskArgs): The arguments for the task.
            context (dict): The context in which to execute the task.
        """
        norm_args = (self.node.type_conversions[i](a) for i, a in enumerate(task))
        return self.node.func(*norm_args, *self.context)
